/* include/lttng/ust-version.h.  Generated from ust-version.h.in by configure.  */
/*
 * SPDX-License-Identifier: MIT
 *
 * Copyright (C) 2011 Mathieu Desnoyers <mathieu.desnoyers@efficios.com>
 *
 * Contains the LTTng-UST versions.
 */

#ifndef _LTTNG_UST_VERSION_H
#define _LTTNG_UST_VERSION_H

#define LTTNG_UST_MAJOR_VERSION 2
#define LTTNG_UST_MINOR_VERSION 13
#define LTTNG_UST_PATCHLEVEL_VERSION 8

#define LTTNG_UST_VERSION "2.13.8"

#define LTTNG_UST_LIB_SONAME_MAJOR 1
#define LTTNG_UST_CTL_LIB_SONAME_MAJOR 5

#endif  /* _LTTNG_UST_VERSION_H */
